/*
   list.c
   sample linear list DLL
*/

/* includes */

#include <stdlib.h>

#ifndef DLLCC
#include <dllif.h>
#endif

/* error numbers */

#define E_HDL   2000    /* no more lists available */
#define E_NOMEM 2001    /* memory allocation failed */
#define E_INVAL 2002    /* invalid list handle */
#define E_NOCUR 2003    /* no current element */
#define E_SIZE  2004    /* string buffer too small */


/* types */

#define MAX_HDL 10

struct list {
   char *name;
   struct entry *first, *last, *curr;
};

struct entry {
   struct entry *next, *prev;
   int len;
   void *data;
};


/* globals */

struct list *hdl_tab[MAX_HDL];


/* 
   cleanup
*/

void dll_cleanup()
{
   int i, hdl;

   for(i = 0; i < MAX_HDL; i++)
      if(hdl_tab[i] != NULL) {
         hdl = i+1;
         (void) l_drop(&hdl);
      }
}


/*
   create list
   CALL DLL List("l_creat", Hdl, "NameOfTheList")
   Hdl will be set to list id (nonzero).
*/

#ifdef __STDC__
EqDLL l_creat( EqInt *hdl, EqString *name )
#else
EqDLL l_creat( hdl, name )
EqInt *hdl;
EqString *name;
#endif
{
   struct list *list;
   int slot;

   /* invalidate hdl */

   *hdl = 0;

   /* find available slot */
      
   for(slot = 0; slot < MAX_HDL; slot++)
      if(hdl_tab[slot] == NULL)
         break;
   if(slot == MAX_HDL)
      return(E_HDL);

   /* allocate list header */
   
   list = (struct list *)calloc(1, sizeof(struct list));
   if(list == NULL)
      return(E_NOMEM);

   /* allocate list name */
   
   list->name = malloc(name->l+1);
   if(list->name == NULL) {
      free(list);
      return(E_NOMEM);
   }

   /* setup list header */
   
   hdl_tab[slot] = list;
   memcpy(list->name,  name->p, name->l);
   list->name[name->l] = '\0';
   *hdl = slot + 1;
   return(0);
}


/*
   return name of list
   CALL DLL List("l_name",Hdl, Name$)
   Name$ will return the name of the list.
*/

#ifdef __STDC__
EqDLL l_name( EqInt hdl, EqString *name )
#else
EqDLL l_name( hdl, name )
EqInt hdl;
EqString *name;
#endif
{
   struct list *list;
   int len;
   
   /* get list header */

   if(hdl < 1 || hdl > MAX_HDL)
      return(E_INVAL);
   if(hdl_tab[ hdl-1 ] == NULL)
      return(E_INVAL);
   
   list = hdl_tab[ hdl-1 ];

   /* return name */

   len = strlen(list->name);
   if(len > name->maxl)
      return(E_SIZE);

   name->l = len;
   strcpy(name->p, list->name);
   return(0);
}


/*
   drop the list (and all elements)
   CALL DLL List("l_drop", Hdl)
   Hdl will be set to zero.
*/

#ifdef __STDC__
EqDLL l_drop( EqInt *hdl )
#else
EqDLL l_drop( hdl )
EqInt *hdl;
#endif
{
   struct list *list;
   struct entry *entry;

   /* get list header */

   if(*hdl < 1 || *hdl > MAX_HDL)
      return(E_INVAL);
   if(hdl_tab[ *hdl-1 ] == NULL)
      return(E_INVAL);
   
   list = hdl_tab[ *hdl-1 ];

   /* drop entries */
   
   entry = list->first;
   while(entry) {
      if(entry->len)
         free(entry->data);
      free(entry);
      entry = entry->next;
   }

   /* drop list */

   free(list->name);
   free(list);
   hdl_tab[ *hdl-1 ] = NULL;
   *hdl = 0;
   return(0);
}


/*
   add element to list
   CALL DLL List("l_add", Mode, Data$)

   Mode: 0 - append to list
         1 - insert before current element, make current
*/

#ifdef __STDC__
EqDLL l_add( EqInt hdl, EqInt mode, EqString *data )
#else
EqDLL l_add( hdl, mode, data )
EqInt hdl, mode;
EqString *data;
#endif
{
   struct list *list;
   struct entry *entry;
   
   /* get list header */

   if(hdl < 1 || hdl > MAX_HDL)
      return(E_INVAL);
   if(hdl_tab[ hdl-1 ] == NULL)
      return(E_INVAL);
   
   list = hdl_tab[ hdl-1 ];

   /* allocate entry */

   entry = calloc(1, sizeof(struct entry));
   if(entry == NULL)
      return(E_NOMEM);

   entry->len = data->l;
   if(data->l) {
      entry->data = malloc(data->l);
      if(entry->data == NULL) {
         free(entry);
         return(E_NOMEM);
      }
      memcpy(entry->data, data->p, data->l);
   }

   /* link entry into list */

   if(mode == 0)
   {
      /* append */

      if(list->last == NULL)
         list->first = list->last = entry;
      else {
         list->last->next = entry;
         entry->prev = list->last;
         list->last = entry;
      }
   }
   else {

      /* insert current */

      if(list->curr == NULL)
         return(E_NOCUR);

      entry->prev = list->curr->prev;
      entry->next = list->curr;

      if(entry->prev == NULL)
         list->first = entry;
      else
         entry->prev->next = entry;

      if(entry->next == NULL)
         list->last = entry;
      else
         entry->next->prev = entry;
      list->curr = entry;
   }
   return(0);
}


/*
   delete current element
   CALL DLL List("l_hdl", Eol)
   Eol will indicate the end-of-list condition.
*/

#ifdef __STDC__
EqDLL l_del( EqInt hdl, EqInt *eol )
#else
EqDLL l_del( hdl, eol )
EqInt hdl, *eol;
#endif
{
   struct list *list;
   
   /* get list header */

   if(hdl < 1 || hdl > MAX_HDL)
      return(E_INVAL);
   if(hdl_tab[ hdl-1 ] == NULL)
      return(E_INVAL);
   
   list = hdl_tab[ hdl-1 ];

   /* unlink from list */

   if(list->curr == NULL)
      return(E_NOCUR);
   
   if(list->curr->prev == NULL)
      list->first = list->curr->next;
   else
      list->curr->prev->next = list->curr->next;

   if(list->curr->next == NULL)
      list->last = list->curr->prev;
   else
      list->curr->next->prev = list->curr->prev;

   /* free entry */
   
   if(list->curr->len)
      free(list->curr->data);
   free(list->curr);
   list->curr = list->curr->next;
   *eol = list->curr == NULL;
   return(0);
}


/*
   get list element
   CALL DLL List("l_get", Mode, Data$, Eol)
   Data$ will return element value.
   Eol will indicate the end-of-list condition.
*/

#ifdef __STDC__
EqDLL l_get( EqInt hdl, EqInt mode, EqString *data, EqInt *eol)
#else
EqDLL l_get( hdl, mode, data, eol)
EqInt hdl, mode, *eol;
EqString *data;
#endif
{
   struct list *list;
   
   /* get list header */

   if(hdl < 1 || hdl > MAX_HDL)
      return(E_INVAL);
   if(hdl_tab[ hdl-1 ] == NULL)
      return(E_INVAL);
   
   list = hdl_tab[ hdl-1 ];

   /* position current */

   switch(mode) {
   case 0:           /* first */
      list->curr = list->first;
      break;
   case 1:           /* last */
      list->curr = list->last;
      break;
   case 2:           /* next */
      if(list->curr == NULL)
         return(E_NOCUR);
      list->curr = list->curr->next;
      break;
   case 3:           /* prev */
      if(list->curr == NULL)
         return(E_NOCUR);
      list->curr = list->curr->prev;
      break;
   }

   *eol = list->curr == NULL;
   if(list->curr) {
      if(list->curr->len > data->maxl)
         return(E_SIZE);
      data->l = list->curr->len;
      if(data->l)
         memcpy(data->p, list->curr->data, data->l);
   }
   return(0);
}

